%% 15-317 Fall 2016 - Prolog Programming
%% October, 25th

%% Eliminate duplicate elements of a list

dedup([], []).
dedup([X], [X]).
dedup([X1,X2|Xs], Ys) :- X1 = X2, dedup([X1|Xs], Ys).
% Could also be written as:
% dedup([X,X|Xs], Ys) :- dedup([X|Xs], Ys).
dedup([X1,X2|Xs], [X1|Ys]) :- X1 \= X2, dedup([X2|Xs], Ys).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% Testing

% Accumulating in the body, does not work.
copy([], []).
copy([X|Xs], [X|Ys]) :- copy(Xs, Ys).
% copy([X|Xs], Ys) :- copy(Xs, [X|Ys]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% Graphs: find a path between two edges

edge(a,b).
edge(b,c).
edge(c,a).
edge(d,b).

% Try edges both ways so we don't need the commutativity clause:
% edge(X, Y) :- edge(Y, X)

path(X, X, _, [X]).
path(X, Y, V, [X|P]) :- X \= Y, not_member(X, V), edge(X, Z), path(Z, Y, [X|V], P).
path(X, Y, V, [X|P]) :- X \= Y, not_member(X, V), edge(Z, X), path(Z, Y, [X|V], P).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% SUDOKU
%
% Board representation
%
% +----------+---------+---------+
% |          |         |         |
% |    A     |    B    |    C    |
% |          |         |         |
% +----------+---------+---------+
% |          |         |         |
% |    D     |    E    |    F    |
% |          |         |         |
% +----------+---------+---------+
% |          |         |         |
% |    G     |    H    |    I    |
% |          |         |         |
% +----------+---------+---------+
%
% [
%   [_,_,_,_,_,_,_,_,_], 
%   [_,_,_,_,_,_,_,_,_], 
%   [_,_,_,_,_,_,_,_,_], 
%
%   [_,_,_,_,_,_,_,_,_], 
%   [_,_,_,_,_,_,_,_,_], 
%   [_,_,_,_,_,_,_,_,_], 
%
%   [_,_,_,_,_,_,_,_,_], 
%   [_,_,_,_,_,_,_,_,_], 
%   [_,_,_,_,_,_,_,_,_]  
% ]

% Unfortunately this solution for cols_ok does not work. It will succeed even if
% only one column is correct. This is the reason why prolog was giving 9 answers,
% one for each column.
%
% get_column(_, [], []).
% get_column(N, [L|M], [X|C]) :- 
%   nth(N, L, X), get_column(N, M, C).
%
% cols_ok(M) :-
%   nth(N, M, _),
%   get_column(N, M, C),
%   permutation([1,2,3,4,5,6,7,8,9], C).

% Here is a recursive solution that looks at each column at a time and recurses
% on the rest of the matrix. This is analogous to what it is done on rows_ok,
% but since a column is not readily available, we need the 'get_firsts' helper.

get_firsts([], [], []).
get_firsts([[X]|M], [X|Fs], Rs) :- get_firsts(M, Fs, Rs).
get_firsts([[X|Xs]|M], [X|Fs], [Xs|Rs]) :- get_firsts(M, Fs, Rs).

cols_ok([]).
cols_ok([H|M]) :-
  get_firsts([H|M], C, Rs),
  permutation([1,2,3,4,5,6,7,8,9], C),
  cols_ok(Rs).

rows_ok([]).
rows_ok([L|M]) :- 
  permutation([1,2,3,4,5,6,7,8,9], L),
  rows_ok(M).

boxs_ok([]).
boxs_ok([[],[],[]|M]) :- boxs_ok(M).
boxs_ok([[X1,X2,X3|L1],
         [Y1,Y2,Y3|L2],
         [Z1,Z2,Z3|L3]|M]) :-
  B = [X1,X2,X3,Y1,Y2,Y3,Z1,Z2,Z3],
  permutation([1,2,3,4,5,6,7,8,9], B),
  boxs_ok([L1,L2,L3|M]).  

sudoku(M) :- 
  rows_ok(M),
  cols_ok(M),
  boxs_ok(M).

%% Tests:
%
% Complete board
%
% sudoku([[9,3,4,8,2,5,6,1,7],
% [6,7,2,9,1,4,8,5,3],
% [5,1,8,6,3,7,9,2,4],
% [3,2,5,7,4,8,1,6,9],
% [4,6,9,1,5,3,7,8,2],
% [7,8,1,2,6,9,4,3,5],
% [1,9,7,5,8,2,3,4,6],
% [8,5,3,4,7,6,2,9,1],
% [2,4,6,3,9,1,5,7,8]]).
%
% Partially empty board (takes a veeeery long time)
%
% sudoku([[A9,A3,4,8,A2,A5,A6,1,7],
% [6,7,B2,9,B1,B4,B8,B5,B3],
% [5,C1,8,C6,3,C7,C9,C2,4],
% [3,D2,D5,7,4,D8,1,D6,D9],
% [E4,6,9,E1,E5,E3,7,8,E2],
% [F7,F8,1,F2,6,9,F4,F3,5],
% [1,G9,G7,G5,8,G2,3,G4,6],
% [H8,H5,H3,H4,H7,6,H2,9,1],
% [2,4,I6,I3,I9,1,5,I7,I8]]).

